use anyhow::{anyhow, Result};
use chrono::{DateTime, Utc};
use sqlx::FromRow;

use crate::repo::Storage;

impl Storage {
    #[tracing::instrument(skip(self), name = "Storage::retrieve_dialog")]
    pub async fn retrieve_dialog(&self, participant_1: &str, participant_2: &str) -> Result<Option<StoredDialog>> {
        let mut executor = self.conn.lock().await;
        let res = sqlx::query_as(
            "select r.id, r.participant_1, r.participant_2, r.message_count
            from dialogs r join users u1 on r.participant_1 = u1.id join users u2 on r.participant_2 = u2.id
            where u1.name = ? and u2.name = ?;",
        )
        .bind(participant_1)
        .bind(participant_2)
        .fetch_optional(&mut *executor)
        .await?;

        Ok(res)
    }

    #[tracing::instrument(skip(self, content, created_at), name = "Storage::insert_dialog_message")]
    pub async fn insert_dialog_message(
        &self,
        dialog_id: u32,
        id: u32,
        author_id: &str,
        content: &str,
        created_at: &DateTime<Utc>,
    ) -> Result<()> {
        let mut executor = self.conn.lock().await;
        let res: Option<(u32,)> = sqlx::query_as("select id from users where name = ?;")
            .bind(author_id)
            .fetch_optional(&mut *executor)
            .await?;
        let Some((author_id,)) = res else {
            return Err(anyhow!("No such user"));
        };
        sqlx::query(
            "insert into dialog_messages(dialog_id, id, author_id, content, created_at)
            values (?, ?, ?, ?, ?);
            update dialogs set message_count = message_count + 1 where id = ?;",
        )
        .bind(dialog_id)
        .bind(id)
        .bind(author_id)
        .bind(content)
        .bind(created_at)
        .bind(dialog_id)
        .execute(&mut *executor)
        .await?;

        Ok(())
    }

    #[tracing::instrument(skip(self, created_at), name = "Storage::initialize_dialog")]
    pub async fn initialize_dialog(
        &self,
        participant_1: &str,
        participant_2: &str,
        created_at: &DateTime<Utc>,
    ) -> Result<StoredDialog> {
        let mut executor = self.conn.lock().await;
        let res: StoredDialog = sqlx::query_as(
            "insert into dialogs(participant_1, participant_2, created_at)
            values (
                (select id from users where name = ?),
                (select id from users where name = ?),
                ?
            )
            returning id, participant_1, participant_2, message_count;",
        )
        .bind(participant_1)
        .bind(participant_2)
        .bind(&created_at)
        .fetch_one(&mut *executor)
        .await?;

        Ok(res)
    }
}

#[derive(FromRow)]
pub struct StoredDialog {
    pub id: u32,
    pub participant_1: u32,
    pub participant_2: u32,
    pub message_count: u32,
}
